"""Developer API.

Paths and file system:

.. autosummary::
   :toctree:

   upath

Settings:

.. autosummary::
   :toctree:

   UserSettings
   InstanceSettings
   StorageSettings

"""
from . import upath
from ._deprecated import deprecated
from ._docs import doc_args
from ._settings_instance import InstanceSettings
from ._settings_user import UserSettings
from ._storage import StorageSettings
from ._testdb import setup_local_test_postgres, setup_local_test_sqlite_file

# below is for backward compat
Storage = StorageSettings



from lamin_logger import logger

from .dev._settings_store import current_instance_settings_file


def check_instance_setup(from_lamindb: bool = False):
    if current_instance_settings_file().exists():
        try:
            # attempt loading the settings file
            from .dev._settings_load import load_instance_settings

            load_instance_settings()

            # if importing from lamindb, also ensure migrations are correct
            if from_lamindb:
                # attempt accessing settings and migrating the instance
                from . import settings
                from ._migrate import check_deploy_migration

                check_deploy_migration(
                    usettings=settings.user, isettings=settings.instance
                )
            # set the check to true
            return True
        except Exception:
            # user will get more detailed traceback once they run the CLI
            raise RuntimeError(
                "Your current instance cannot be reached.\nInit or load a connectable"
                " instance on the command line: `lamin load <instance>` or `lamin init"
                " <...>`"
            )
    else:
        logger.warning(
            "You haven't yet setup an instance using the CLI: Please call"
            " `ln.setup.init()` or `ln.setup.load()`"
        )
        return False



import os
import shutil
from pathlib import Path

from lamin_logger import logger
from pydantic import BaseSettings


def get_settings_dir():
    settings_dir = Path.home() / ".lamin"
    settings_dir.mkdir(parents=True, exist_ok=True)
    # deal with legacy settings directory
    legacy_dir = settings_dir.with_name(".lndb")
    if legacy_dir.exists():
        if not settings_dir.exists():
            legacy_dir.rename(settings_dir)
            logger.info(f"Renamed legacy settings dir {legacy_dir} to {settings_dir}")
        else:
            for path in legacy_dir.glob("*"):
                shutil.copy(path, settings_dir)
            logger.info(
                f"Copied content of legacy settings dir {legacy_dir} to {settings_dir}."
                f" You can delete {legacy_dir}!"
            )
    return settings_dir


# user_config_dir in appdirs is weird on MacOS!
# hence, let's take home/.lndb
settings_dir = get_settings_dir()


def get_settings_file_name_prefix():
    if "LAMIN_ENV" in os.environ:
        if os.environ["LAMIN_ENV"] == "staging":
            return "staging-"
    return ""


def current_instance_settings_file():
    return settings_dir / f"{get_settings_file_name_prefix()}current_instance.env"


def current_user_settings_file():
    return settings_dir / f"{get_settings_file_name_prefix()}current_user.env"


def instance_settings_file(name: str, owner: str):
    return (
        settings_dir / f"{get_settings_file_name_prefix()}{owner}-instance-{name}.env"
    )


def user_settings_file_email(email: str):
    return settings_dir / f"{get_settings_file_name_prefix()}user-{email}.env"


def user_settings_file_handle(handle: str):
    return settings_dir / f"{get_settings_file_name_prefix()}user-{handle}.env"


class InstanceSettingsStore(BaseSettings):
    owner: str
    name: str
    storage_root: str
    storage_region: str  # should be Optional, but we use types for instantiating
    db: str  # should be Optional, but we use types for instantiating
    schema_str: str  # should be Optional, but we use types for instantiating

    class Config:
        env_file = ".env"


class UserSettingsStore(BaseSettings):
    email: str
    password: str
    access_token: str
    id: str
    handle: str
    name: str

    class Config:
        env_file = ".env"



from pathlib import Path
from typing import Optional

from pydantic.error_wrappers import ValidationError

from ._settings_instance import InstanceSettings
from ._settings_store import (
    InstanceSettingsStore,
    UserSettingsStore,
    current_instance_settings_file,
    current_user_settings_file,
)
from ._settings_user import UserSettings


def load_instance_settings(instance_settings_file: Optional[Path] = None):
    if instance_settings_file is None:
        instance_settings_file = current_instance_settings_file()
    if not instance_settings_file.exists():
        raise RuntimeError("No instance is loaded! Call `lamin init` or `lamin load`")
    try:
        settings_store = InstanceSettingsStore(_env_file=instance_settings_file)
    except ValidationError:
        raise RuntimeError(
            "Your instance settings file is invalid, please delete"
            f" {instance_settings_file} and init the instance again."
        )
    isettings = setup_instance_from_store(settings_store)
    return isettings


def load_or_create_user_settings():
    """Return current user settings."""
    if not current_user_settings_file().exists():
        global UserSettings
        return UserSettings()
    else:
        settings = load_user_settings(current_user_settings_file())
        return settings


def load_user_settings(user_settings_file: Path):
    try:
        settings_store = UserSettingsStore(_env_file=user_settings_file)
    except ValidationError:
        raise RuntimeError(
            "Your user settings file is invalid, please delete"
            f" {user_settings_file} and log in again."
        )
    settings = setup_user_from_store(settings_store)
    return settings


def setup_instance_from_store(store: InstanceSettingsStore) -> InstanceSettings:
    return InstanceSettings(
        owner=store.owner,
        name=store.name,
        storage_root=store.storage_root,
        storage_region=store.storage_region if store.storage_region != "null" else None,
        db=store.db if store.db != "null" else None,  # type: ignore
        schema=store.schema_str if store.schema_str != "null" else None,
    )


def setup_user_from_store(store: UserSettingsStore) -> UserSettings:
    settings = UserSettings()
    settings.email = store.email
    settings.password = store.password if store.password != "null" else None
    settings.access_token = store.access_token
    settings.id = store.id if store.id != "null" else None
    settings.handle = store.handle if store.handle != "null" else None
    settings.name = store.name if store.name != "null" else None
    return settings



import sqlalchemy as sql

from ._settings import settings


class schema:
    """Inspect the schema."""

    @classmethod
    def draw(cls, view=True):
        """Make a diagram of entity relationships."""
        import erdiagram

        engine = settings.instance.engine
        metadata = sql.MetaData(bind=engine)
        metadata.reflect()
        graph = erdiagram.create_schema_graph(
            metadata=metadata,
            show_datatypes=False,
            show_indexes=False,  # ditto for indexes
            rankdir="TB",
            concentrate=False,  # Don't try to join the relation lines together
        )
        if view:
            erdiagram.view(graph)
        else:
            return graph

    @classmethod
    def list_entities(cls):
        """Return all entities in the db."""
        metadata = sql.MetaData()
        engine = settings.instance.engine
        metadata.reflect(bind=engine)
        table_names = [table.name for table in metadata.sorted_tables]
        return table_names



import os

from lamin_logger import logger

from ._settings import settings
from .dev._settings_store import current_instance_settings_file


def close() -> None:
    """Close existing instance.

    Returns `None` if succeeds, otherwise an exception is raised.
    """
    if current_instance_settings_file().exists():
        instance = settings.instance.identifier
        current_instance_settings_file().unlink()
        os.environ["LAMINDB_INSTANCE_LOADED"] = "0"
        logger.success(f"Closed {instance}")
    else:
        logger.info("No instance loaded")



import argparse
import os
import sys

from lamin_logger import logger

from . import _init_instance, _setup_user, delete, info, set_storage
from ._close import close as close_instance
from ._init_instance import description as instance
from .dev._settings_user import user_description as user

signup_help = "First time sign up."
login_help = "Log in an already-signed-up user."
init_help = "Init & config instance with db & storage."
load_help = "Load instance by name."
set_storage_help = "Set storage used by an instance."
info_help = "Show current instance information."
close_help = "Close instance."
migr_help = "Manage migrations."
delete_help = "Delete an instance."

description_cli = "Configure LaminDB and perform simple actions."
parser = argparse.ArgumentParser(
    description=description_cli, formatter_class=argparse.RawTextHelpFormatter
)
subparsers = parser.add_subparsers(dest="command")

# signup user
signup = subparsers.add_parser("signup", help=signup_help)
aa = signup.add_argument
aa("email", type=str, metavar="email", help=user.email)

# login user
login = subparsers.add_parser("login", help=login_help)
aa = login.add_argument
aa(
    "user",
    type=str,
    metavar="user",
    help="Email or user handle. Email is needed at first login.",
)  # noqa
aa("--password", type=str, metavar="s", default=None, help=user.password)

# init instance
init = subparsers.add_parser("init", help=init_help)
aa = init.add_argument
aa("--storage", type=str, metavar="s", help=instance.storage_root)
aa("--db", type=str, metavar="s", default=None, help=instance.db)
aa("--schema", type=str, metavar="s", default=None, help=instance.schema)
aa("--name", type=str, metavar="s", default=None, help=instance.name)

# load instance
load = subparsers.add_parser("load", help=load_help)
aa = load.add_argument
instance_help = """
The instance identifier can the instance name (owner is
current user), handle/name, or the URL: https://lamin.ai/handle/name."""
aa("instance", type=str, metavar="s", default=None, help=instance_help)

# delete instance
delete_parser = subparsers.add_parser("delete", help=delete_help)
aa = delete_parser.add_argument
aa("instance", type=str, metavar="s", default=None, help=instance.name)

# show instance info
info_parser = subparsers.add_parser("info", help=info_help)

# set storage
set_storage_parser = subparsers.add_parser("set", help=set_storage_help)
aa = set_storage_parser.add_argument
aa("--storage", type=str, metavar="s", help=instance.storage_root)

# close instance
close = subparsers.add_parser("close", help=close_help)

# migrate
migr = subparsers.add_parser("migrate", help=migr_help)
aa = migr.add_argument
aa("action", choices=["generate"], help="Generate migration.")

# parse args
args = parser.parse_args()


def process_result(result):
    if result in ["migrate-unnecessary", "migrate-success", "migrate-skipped", None]:
        return None  # is interpreted as success (exit code 0) by shell
    else:
        return result


def main():
    if os.path.basename(sys.argv[0]) == "lndb":
        logger.warning("CLI `lndb` is deprecated, use `lamin` instead.")
    if args.command == "signup":
        return _setup_user.signup(email=args.email)
    if args.command == "login":
        return _setup_user.login(
            args.user,
            password=args.password,
        )
    elif args.command == "init":
        result = _init_instance.init(
            storage=args.storage, db=args.db, schema=args.schema, name=args.name
        )
        return process_result(result)
    elif args.command == "load":
        result = _init_instance.load(
            identifier=args.instance,
        )
        return process_result(result)
    elif args.command == "close":
        return close_instance()
    elif args.command == "delete":
        return delete(
            instance_name=args.instance,
        )
    elif args.command == "info":
        return info()
    elif args.command == "set":
        return set_storage(storage=args.storage)
    elif args.command == "migrate":
        from . import migrate

        if args.action == "generate":
            return migrate.generate()
    else:
        logger.error("Invalid command. Try `lamin -h`.")
        return 1



"""Paths & file systems.

.. autosummary::
   :toctree:

   UPath
   infer_filesystem

"""

import os
from datetime import timezone
from pathlib import Path
from typing import Union

import fsspec
from dateutil.parser import isoparse  # type: ignore
from lamin_logger import logger
from upath import UPath


def infer_filesystem(path: Union[Path, UPath, str]):
    path_str = str(path)

    if isinstance(path, UPath):
        fs = path.fs
    else:
        protocol = fsspec.utils.get_protocol(path_str)
        if protocol == "s3":
            fs_kwargs = {"cache_regions": True}
        else:
            fs_kwargs = {}
        fs = fsspec.filesystem(protocol, **fs_kwargs)

    return fs, path_str


def _download_to(self, path, **kwargs):
    self.fs.download(str(self), str(path), **kwargs)


def _upload_from(self, path, **kwargs):
    self.fs.upload(str(path), str(self), **kwargs)


def _synchronize(self, filepath: Path):
    if not self.exists():
        return None

    if not filepath.exists():
        filepath.parent.mkdir(parents=True, exist_ok=True)
        mts = self.modified.timestamp()
        self.download_to(filepath)
        os.utime(filepath, times=(mts, mts))
        return None

    cloud_mts = self.modified.timestamp()
    local_mts = filepath.stat().st_mtime
    if cloud_mts > local_mts:
        mts = self.modified.timestamp()
        self.download_to(filepath)
        os.utime(filepath, times=(mts, mts))
    elif cloud_mts < local_mts:
        logger.warning(
            f"Local file ({filepath}) for cloud path ({self}) is newer on disk than in cloud.\n"  # noqa
            "It seems you manually updated the database locally and didn't push changes to the cloud.\n"  # noqa
            "This can lead to data loss if somebody else modified the cloud file in"  # noqa
            " the meantime."
        )


def _modified(self):
    path = str(self)
    if "gcs" not in self.fs.protocol:
        mtime = self.fs.modified(path)
    else:
        stat = self.fs.stat(path)
        if "updated" in stat:
            mtime = stat["updated"]
            mtime = isoparse(mtime)
        else:
            return None
    # always convert to the local timezone before returning
    # assume in utc if the time zone is not specified
    if mtime.tzinfo is None:
        mtime = mtime.replace(tzinfo=timezone.utc)
    return mtime.astimezone().replace(tzinfo=None)


UPath.download_to = _download_to
UPath.upload_from = _upload_from
UPath.synchronize = _synchronize
UPath.modified = property(_modified)



from pathlib import Path
from typing import Union

from lamin_logger import logger
from lnhub_rest._add_storage import add_storage as add_storage_hub

from lndb.dev.upath import UPath

from ._init_instance import register
from ._settings import settings
from .dev import deprecated
from .dev._settings_instance import InstanceSettings


class set:
    """Set properties of current instance."""

    @staticmethod
    def storage(root: Union[str, Path, UPath]):
        """Set storage."""
        if settings.instance.owner != settings.user.handle:
            logger.error("Can only set storage if current user is instance owner.")
            return "only-owner-can-set-storage"

        if settings.instance.dialect == "sqlite":
            logger.error("Can't set storage for sqlite instances.")
            return "set-storage-failed"

        new_isettings = InstanceSettings(
            owner=settings.instance.owner,
            name=settings.instance.name,
            storage_root=root,
            db=settings.instance.db,
            schema=settings.instance._schema_str,
        )

        new_isettings._persist()
        register(new_isettings, settings.user)
        if settings.instance.is_remote:
            add_storage_hub(root, account_handle=settings.instance.owner)

        logger.info(f"Set storage {root}")


@deprecated("lndb.set.storage()")
def set_storage(storage: Union[str, Path, UPath]):
    """Deprecated in favor of `set.storage`."""
    set.storage(storage)
